import os

import pandas as pd
from decide import data_folder
from decide.data.database import connection, Manager

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)


def write_result(conn, model_run_id, output_directory):

    df = pd.read_sql("""
    SELECT  
        a.p as p,
        a.issue as issue,  
        a.repetion || '-' || a.iteration as pointer, 
        a.numerator / a.denominator AS nbs
    FROM (SELECT
            sum(ai.position * ai.power * ai.salience) AS numerator,
            sum(ai.salience * ai.power)               AS denominator,
            r.pointer                                 AS repetion,
            i2.pointer                                AS iteration,
            m.p,
            i.name as issue
          FROM actorissue ai
            LEFT JOIN issue i ON ai.issue_id = i.id
            LEFT JOIN actor a ON ai.actor_id = a.id
            LEFT JOIN iteration i2 ON ai.iteration_id = i2.id
            LEFT JOIN repetition r ON i2.repetition_id = r.id
            LEFT JOIN modelrun m ON r.model_run_id = m.id            
          WHERE  ai.type = 'before' AND i2.pointer = 9 AND m.id = ? 
         GROUP BY m.id,r.id, i2.id, i.id) a
    """,
                     conn,
                     params=(model_run_id, ),
                     index_col=['p'],
                     columns=['issue']
                     )

    for p in sorted(set(df.index)):
        x = df.loc[p].pivot(index='pointer', columns='issue', values='nbs').cov().round(5)

        x.to_csv(os.path.join(output_directory, 'covariance_{}.csv'.format(p)))

        print('writen covariance table for p={}'.format(p))


if __name__ == '__main__':
    from environs import Env
    env = Env()
    Env.read_env()  # read .env file, if it exists

    url = env('DATABASE_URL')

    m = Manager(url)
    m.init_database()

    model_run_id = 1

    write_result(connection, model_run_id, data_folder)



import uuid
from decimal import Decimal

from PyQt5 import QtWidgets
from typesystem.base import ValidationResult

from decide.data import types
from decide.qt.inputwindow import signals
from decide.qt.mainwindow.helpers import DoubleInput


class BaseInputModel:
    type = None

    def __init__(self):
        self.elements = {}
        self.uuid = uuid.uuid4()
        self.validation_result = ValidationResult()
        self.widgets = {}
        self.stylesheets = {}

    def is_valid(self):
        self.reset_state()
        self.validation_result = self.type.validate_or_error(self.as_dict())

        if self.validation_result.error:
            self.handle_error()
            return False
        else:
            return True

    def as_dict(self):
        raise NotImplementedError

    def handle_error(self):
        for key, value in self.validation_result.error.items():
            if key in self.widgets:

                if key not in self.stylesheets:
                    self.stylesheets[key] = self.widgets[key].styleSheet()

                self.widgets[key].setStyleSheet("border: 1px solid red")
                self.widgets[key].setToolTip(self.validation_result.error[key])

    def reset_state(self):
        for key, value in self.stylesheets.items():
            self.widgets[key].setStyleSheet(value)


class ActorInputModel(BaseInputModel):
    """
    Object containing a name and power
    """

    type = types.Actor

    def __init__(self, name: str, power: Decimal):
        super().__init__()
        self.id = None
        self._name = name
        self._power = power
        self._comment = ""
        self.key = "actor_input"

    def as_dict(self):
        return {
            'id': self.name,
            'power': self.power,
            'comment': self.comment
        }

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self.set_name(value)

    @property
    def power(self):
        return self._power

    @power.setter
    def power(self, value):
        self.set_power(value)

    def set_name(self, value, silence=False):
        self._name = value

        if self.is_valid() and not silence:
            signals.actor_changed.send(self, key="name", value=value)

    def set_power(self, value, silence=False):
        self._power = value

        if self.is_valid() and not silence:
            signals.actor_changed.send(self, key="power", value=value)

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, value):
        self.set_comment(value)

    def set_comment(self, value):
        self._comment = value


class IssueInputModel(BaseInputModel):
    """
    Object containing a name, lower and upper bounds
    """

    type = types.Issue

    def __init__(self, name: str, lower: Decimal, upper: Decimal):
        super().__init__()
        self.id = None
        self._name = name
        self._lower = lower
        self._upper = upper
        self._comment = ""
        self.key = "issue_input"

    def as_dict(self):
        return {
            'name': self.name,
            'lower': self.lower,
            'upper': self.upper,
            'comment': self.comment
        }

    @property
    def name(self):
        return self._name

    @property
    def lower(self):
        return self._lower

    @property
    def upper(self):
        return self._upper

    @name.setter
    def name(self, value):
        self.set_name(value)

    @lower.setter
    def lower(self, value):
        self.set_lower(value)

    @upper.setter
    def upper(self, value):
        self.set_upper(value)

    def set_name(self, value, silence=False):
        self._name = value

        if self.is_valid() and not silence:
            signals.issue_changed.send(self, key="name", value=value)

    def set_lower(self, value, silence=False):

        self._lower = value

        if self.is_valid() and not silence:
            signals.issue_changed.send(self, key="lower", value=value)

    def set_upper(self, value, silence=False):

        self._upper = value

        if self.is_valid() and not silence:
            signals.issue_changed.send(self, key="upper", value=value)

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, value):
        self.set_comment(value)

    def set_comment(self, value):
        self._comment = value


class ActorIssueInputModel(BaseInputModel):
    type = types.ActorIssue

    def __init__(self, actor: ActorInputModel, issue: IssueInputModel):
        super().__init__()
        self.id = None

        self.actor = actor
        self.issue = issue

        self.actor_input = QtWidgets.QLabel(actor.name)
        self.issue_input = QtWidgets.QLabel(issue.name)

        self.power_input = QtWidgets.QLabel(str(actor.power))

        self.position_input = DoubleInput()
        self.position_input.valueChanged.connect(self.set_position)

        self.salience_input = DoubleInput()

        self.meta = dict()

        self.uuid = uuid.uuid4()

        self._power = Decimal(0.0)
        self._position = Decimal(0.0)
        self._salience = Decimal(0.0)

    def as_dict(self):
        return {
            'position': self.position,
            'salience': self.salience,
            'power': self.power,
            'issue': self.issue.name,
            'actor': self.actor.name,
        }

    @property
    def position(self):
        return self._position

    @property
    def salience(self):
        return self._salience

    @property
    def power(self):
        return self._power

    def set_position(self, value: Decimal, silence=False):
        self._position = value
        if self.is_valid() and not silence:
            signals.actor_issue_changed.send(self, key="position", value=value)

    def set_power(self, value: Decimal, silence=False):
        self._power = value
        if self.is_valid() and not silence:
            signals.actor_issue_changed.send(self, key="power", value=value)

    def set_salience(self, value: Decimal, silence=False):
        self._salience = value
        if self.is_valid() and not silence:
            signals.actor_issue_changed.send(self, key="salience", value=value)



import pytest

from decide.data.database import Manager


@pytest.fixture(scope='session')
def sqlite_db():
    m = Manager('sqlite:///:memory:')
    m.init_database()
    m.create_tables()

    print('initialized test database')

    yield m

    m.delete_tables()

    print('cleaned test database')



import csv
from typing import List

from decide.model.base import AbstractExchange


def write_exchanges(
        filename: str, realized: List[AbstractExchange], delimiter=";", lineterminator="\n"
):
    """
    Write all realized exchanges to the given file
    """

    if len(realized) <= 0:
        return

    with open(filename, "w") as csv_file:
        writer = csv.writer(
            csv_file, delimiter=delimiter, lineterminator=lineterminator
        )

        # write heading
        writer.writerow(realized[0].csv_row(head=True))

        for exchange in realized:
            if exchange.is_valid:
                writer.writerow(exchange.csv_row())



import typesystem
from typesystem import ValidationError

from decide.model import base


class CSVColumn:
    starts_with = None
    comment = typesystem.Text(allow_null=True, allow_blank=True)


class Comment(CSVColumn, typesystem.Schema):
    starts_with = "#/"
    description = typesystem.Text()

    def __hash__(self):
        return hash(self.description)

    def __eq__(self, other):
        if isinstance(other, Comment):
            return self.description == other.description

        return NotImplemented


class IssueDescription(CSVColumn, typesystem.Schema):
    starts_with = "#O"
    description = typesystem.Text()

    def __hash__(self):
        return hash(self.description)

    def __eq__(self, other):
        if isinstance(other, IssueDescription):
            return self.description == other.description

        return NotImplemented


class PartialActor(CSVColumn, typesystem.Schema):
    starts_with = "#A"
    id = typesystem.String()
    fullname = typesystem.String()

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):

        if isinstance(other, str):
            return self.id == other

        if isinstance(other, PartialActor):
            return self.id == other.id

        return NotImplemented

    def __lt__(self, other):
        return self.id < other.id

    def as_model_object(self):

        actor = base.Actor(self.id, self.fullname)

        return actor


class Actor(PartialActor):
    power = typesystem.Float(minimum=0, maximum=1)


class PartialIssue(CSVColumn, typesystem.Schema):
    """
    An issue as referenced in the csv data file
    """
    starts_with = "#P"
    name = typesystem.String()
    description = typesystem.Text(
        allow_blank=True, allow_null=True
    )

    def __init__(self, *args, **kwargs):
        super(PartialIssue, self).__init__(*args, **kwargs)
        self.lower = None
        self.upper = None

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):

        if isinstance(other, str):
            return self.name == other

        if isinstance(other, PartialIssue):
            return self.name == other.name

        return NotImplemented

    def __lt__(self, other):
        return self.name < other.name

    def validate_interval(self):

        if self.lower > self.upper:
            raise ValidationError(key='interval', text='lower needs to be less then upper')

    def as_model_object(self):

        issue = base.Issue(self.name, self.lower, self.upper)

        return issue


class Issue(PartialIssue):
    """
    An issue from the InputWindow
    """
    name = typesystem.String()
    upper = typesystem.Float()
    lower = typesystem.Float()
    comment = typesystem.String(allow_blank=True)


class IssuePosition(CSVColumn, typesystem.Schema):
    """
    #M;Issue ID;position;meaning;
    """

    starts_with = "#M"
    issue = typesystem.String()
    position = typesystem.Number()
    description = typesystem.Text(allow_blank=True, allow_null=True)

    def __hash__(self):
        return hash(self.issue)

    def __eq__(self, other):
        if isinstance(other, PartialIssue):
            return self.issue == other.issue and self.position == other.position

        return NotImplemented


class ActorIssue(CSVColumn, typesystem.Schema):
    """
    #A;actor;issue;position;salience;power
    """

    starts_with = "#D"
    actor = typesystem.String()
    issue = typesystem.String()
    position = typesystem.Decimal()
    salience = typesystem.Decimal(minimum=0, maximum=1)
    power = typesystem.Decimal(minimum=0, maximum=1)
    comment = typesystem.Text(allow_blank=True, allow_null=True)

    def __str__(self):
        return str(self.actor) + "-" + str(self.issue)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):

        if isinstance(other, PartialIssue):
            return self.__str__() == other.__str__()

        return NotImplemented

    def __lt__(self, other):
        return self.issue.__lt__(other.issue)

    def validate_position(self, issue: PartialIssue):

        issue.validate_interval()

        if not issue.lower <= self.position <= issue.upper:
            raise ValidationError(
                key='position',
                text='exceeds the issue interval of {}-{}'.format(issue.lower, issue.upper)
            )



import logging
from typing import List

from .. import base
from .. import equalgain
from .. import randomrate


class Observer(object):
    """
    The main event Object containing all the event methods.

    Order of events:

    before_repetitions()
    before_iterations()
    before_loop()
    after_loop()
    end_loop()
    after_iterations()
    after_repetitions()

    """

    def __init__(self, observable: "Observable"):
        """
        Init
        :param observable: Observable reference to the main event handler
        :param model_ref:  AbstractModel this reference is almost always needed
        """
        observable.register(self)
        self.model_ref = observable.model_ref
        self.output_directory = observable.output_directory

    def update(self, observable, notification_type, **kwargs):
        pass

    def before_loop(self, iteration: int, repetition: int):
        """
        Before the start of a single round.
        The actors are still on there starting position.
        Next event: after_loop
        :param iteration: int the current iteration number
        :param repetition: int only used in the RandRate model, represents the current number of repetition.
        """
        pass

    def after_loop(
        self, realized: List[base.AbstractExchange], iteration: int, repetition: int
    ):
        """
        After all exchanges are executed. There are no potential exchanges left, but the model reference to actor
        issues is still on there starting position.
        Next event: end_loop
        :param realized:
        :param iteration:
        :param repetition:
        :return:
        """
        pass

    def end_loop(self, iteration: int, repetition: int):
        """
        At the end of a loop with the final positions of the actors, but before the new_start_position()'s are calculated.
        The next event: before_loop.
        """
        pass

    def before_iterations(self, repetition):
        """
        Before a set of loops starts
        """
        pass

    def after_iterations(self, repetition):
        """
        After a set of loops are finished
        """
        pass

    def before_repetitions(self, repetitions, iterations):
        """
        First event
        :return:
        """
        pass

    def after_repetitions(self):
        """
        Last event
        """
        pass

    @staticmethod
    def log(message: str):
        logging.info(message)

    def execute_exchange(self, exchange: base.AbstractExchange):
        if isinstance(exchange, equalgain.EqualGainExchange):
            return self._execute_equal_exchange(exchange)
        elif isinstance(exchange, randomrate.RandomRateExchange):
            return self._execute_random_exchange(exchange)

    def _execute_equal_exchange(self, exchange: equalgain.EqualGainExchange):
        pass

    def _execute_random_exchange(self, exchange: randomrate.RandomRateExchange):
        pass

    def removed_exchanges(self, exchanges: List[base.AbstractExchange]):
        pass


class Observable(Observer):
    """
    Even the Observable is and Observer, this gives us the possibility to call the right method with the correct params
    """

    def __init__(self, model_ref: base.AbstractModel, output_directory: str):
        self.__observers = []
        self.model_ref = model_ref
        self.output_directory = output_directory

    def update_model_ref(self, model):
        self.model_ref = model
        for observer in self.__observers:
            observer.model_ref = model

    def update_output_directory(self, output_directory):
        self.output_directory = output_directory

    def register(self, observer):
        self.__observers.append(observer)

    def before_repetitions(self, repetitions, iterations):
        for observer in self.__observers:
            observer.before_repetitions(repetitions, iterations)

    def before_iterations(self, repetition):
        for observer in self.__observers:
            observer.before_iterations(repetition)

    def before_loop(self, iteration: int, repetition: int = None):
        for observer in self.__observers:
            observer.before_loop(iteration, repetition)

    def execute_exchange(self, exchange: base.AbstractExchange):
        for observer in self.__observers:
            observer.execute_exchange(exchange)

    def after_loop(
        self, realized: List[base.AbstractExchange], iteration: int, repetition: int
    ):
        for observer in self.__observers:
            observer.after_loop(realized, iteration, repetition)

    def end_loop(self, iteration: int, repetition: int):
        for observer in self.__observers:
            observer.end_loop(iteration, repetition)

    def after_iterations(self, repetition):
        for observer in self.__observers:
            observer.after_iterations(repetition)

    def after_repetitions(self):
        for observer in self.__observers:
            observer.after_repetitions()



from decide.qt.inputwindow.gui import InputWindow


def test_load_copenhagen(qtbot):
    """
    Load settings from ProgramSettings
    """

    x = InputWindow()

    x.load_copenhagen()

    assert len(x.actor_issue_widget.actors) == 11
    assert len(x.actor_issue_widget.issues) == 7
    assert len(x.actor_issue_widget.actor_issues.values()) == 11, "are store nested"



from blinker import Namespace

namespace = Namespace()

settings_changed = namespace.signal("settings_changed")

actor_selection_changed = namespace.signal("actor_selection_changed")

issue_selection_changed = namespace.signal("issue_selection_changed")



import logging
import os
import subprocess
import sys
import traceback

from PyQt5 import QtWidgets

from decide.qt import app


def open_file_natively(path):
    if sys.platform.startswith("darwin"):
        subprocess.call(("open", path))
    elif os.name == "nt":
        os.startfile(path)
    elif os.name == "posix":
        subprocess.call(("xdg-open", path))


def exception_hook(exctype, ex, _traceback):
    """
    Setting the system exception hook so the exception will be logged and the log file displayed
    """

    tb_text = exception_to_string(ex)
    logging.exception(tb_text)

    from decide.qt.mainwindow.errordialog import ErrorDialog

    error_dialog = ErrorDialog(tb_text, parent=app)
    sys.exit(error_dialog.exec_())


def exception_to_string(ex):
    tb_lines = traceback.format_exception(ex.__class__, ex, ex.__traceback__)
    tb_text = "".join(tb_lines)
    return tb_text


def show_user_error(self, message: str):
    QtWidgets.QMessageBox.about(self, "Input data invalid", str(message))



import logging
import random
import uuid
from collections import defaultdict
from decimal import *
from operator import attrgetter
from typing import List

from . import base
from . import calculations


class RandomRateExchangeActor(base.AbstractExchangeActor):
    """
    Random Rate solution of the model.
    Where the expected utility in de Equal Gain solution are equal,
    the utility here is calculated on a random exchange ratio.
    """

    def __str__(self):
        return "{1} {2} {3:.1f} {4:.1f} ({5:.1f}) {0:.10f} ".format(
            self.eu,
            self.actor.name,
            self.supply.issue,
            self.supply.position,
            self.y,
            self.opposite_actor.demand.position,
        )

    def recalculate(self, delta_eu, increase):

        # we should not need the opposite actor, because the move isn't effecting the other actors position
        # opposite = self.exchange.opposite_actor

        delta_o = delta_eu / self.supply.salience

        # determine in which direction the nbs is moving

        if self.nbs_1 < self.nbs_0:
            delta_o *= -1

        # we can also decrease the shift.
        delta_o = delta_o if increase else delta_o * -1

        self.adjust_position_by_outcome(delta_o, increase=increase)

    def adjust_position_by_outcome(self, delta_o, increase):
        new_outcome = (
            self.nbs_1 + delta_o
        )  # we have to use nbs_1 here, because it is an incremental shift.

        position = calculations.position_by_nbs(
            actor_issues=self.exchange.model.actor_issues[self.supply.issue],
            exchange_actor=self,
            nbs=new_outcome,
            denominator=self.exchange.model.nbs_denominators[self.supply.issue],
        )

        if increase and (
            delta_o < 0
            and position < self.opposite_actor.demand.position
            or delta_o > 0
            and position > self.opposite_actor.demand.position
        ):
            position = self.opposite_actor.demand.position

            adjusted_nbs = calculations.adjusted_nbs(
                actor_issues=self.exchange.model.actor_issues[self.supply.issue],
                updates=self.exchange.updates,
                actor=self.actor,
                new_position=position,
                denominator=self.exchange.model.nbs_denominators[self.supply.issue],
            )

            new_outcome = adjusted_nbs
            delta_o_consumed = abs(new_outcome - self.nbs_1) - delta_o
            delta_o_left = delta_o - delta_o_consumed

            self.opposite_actor.recalculate(delta_o_left, increase=False)

        previous_move = self.moves.pop()

        # calculate the move in the right direction
        if previous_move < 0:
            move = abs(self.supply.position - position) * -1
        else:
            move = abs(self.supply.position - position)

        self.moves.append(move)
        self.exchange.is_valid = self.is_move_valid(move)
        if self.exchange.is_valid:
            self.y = position
            if self.exchange.i.y > 100 or self.exchange.j.y > 100:
                self.exchange.is_valid = False
            if self.exchange.i.y < 0 or self.exchange.j.y < 0:
                self.exchange.is_valid = False
            self.nbs_1 = new_outcome

    def adjust_utility(self, delta_o):

        self.eu += abs(delta_o) * self.supply.salience
        self.opposite_actor.eu -= delta_o


class RandomRateExchange(base.AbstractExchange):
    """
    An exchange for the random rate model
    """

    actor_class = RandomRateExchangeActor
    """ For the factory, so the Abstract know's which type he has to create  """

    def __init__(self, i, j, p, q, m, groups):
        super().__init__(i, j, p, q, m, groups)

        self.i.exchange = self
        self.j.exchange = self
        self.key = uuid.uuid4()

    def to_list(self):
        return [self.i, self.j]

    def calculate(self):
        # first we try to move j to the position of i on issue p
        # we start with the calculation for j

        if not self.re_calc:
            a = float(self.j.s_demand / self.i.s)
            b = float(self.i.s_demand / self.j.s)

            if b > a:
                a, b = b, a

            self.dp = Decimal(random.uniform(a, b))
            self.dq = Decimal(random.uniform(a, b))

        self.i.move = calculations.reverse_move(
            self.model.actor_issues[self.i.supply_issue], self.i, self.dq
        )
        self.j.move = abs(self.i.x_demand - self.j.x)

        if abs(self.i.move) > abs(self.j.x_demand - self.i.x):
            self.dq = calculations.by_absolute_move(
                self.model.actor_issues[self.i.supply_issue], self.i
            )
            self.dp = calculations.by_exchange_ratio(self.i, self.dq)

            self.i.move = abs(self.j.x_demand - self.i.x)
            self.j.move = calculations.reverse_move(
                self.model.actor_issues[self.j.supply_issue], self.j, self.dp
            )

        # this check is only necessary for the smallest exchange,
        # because if the smallest exchange exceeds the limit the larger one will definitely do so

        if self.i.x > self.j.x_demand:
            self.i.move *= -1

        if self.j.x > self.i.x_demand:
            self.j.move *= -1

        self.i.moves.append(self.i.move)
        self.j.moves.append(self.j.move)

        self.i.y = self.i.x + self.i.move
        self.j.y = self.j.x + self.j.move

        self.i.eu = abs(calculations.expected_utility(self.i, self.dq, self.dp))
        self.j.eu = abs(calculations.expected_utility(self.j, self.dp, self.dq))

        b1 = self.i.is_move_valid(self.i.move)
        b2 = self.j.is_move_valid(self.j.move)

        self.is_valid = b1 and b2

        if self.is_valid:
            self.i.check_nbs()
            self.j.check_nbs()

    def __str__(self):
        return "{0}, {1}".format(str(self.i), str(self.j))

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        if item == self.i.actor_name:
            return self.i
        elif item == self.j.actor_name:
            return self.j

        raise Exception("Actor {0} not in exchange".format(item))

    def csv_row(self, head=False):

        if head:
            return [
                # the actors
                "actor",
                "issue",
                "power",
                "sal s/d",
                "start",
                "move",
                "voting",
                "demand",
                "gain",
                "exchange ratio",
                "",
                "actor",
                "issue",
                "power",
                "sal s/d",
                "start",
                "move",
                "voting",
                "demand",
                "gain",
                "exchange ratio",
            ]

        exchange = self

        return [
            # the actors
            exchange.i.actor_name,
            exchange.i.supply_issue,
            exchange.i.c,
            exchange.i.s / exchange.i.s_demand,
            exchange.i.x,
            exchange.i.move,
            exchange.i.y,
            exchange.i.opposite_actor.x_demand,
            exchange.i.eu,
            exchange.dp,
            "",
            exchange.j.actor_name,
            exchange.j.supply_issue,
            exchange.j.c,
            exchange.j.s / exchange.j.s_demand,
            exchange.j.x,
            exchange.j.move,
            exchange.j.y,
            exchange.j.opposite_actor.x_demand,
            exchange.j.eu,
            exchange.dq,
        ]


class RandomRateModel(base.AbstractModel):
    """
    The Random Rate implementation
    """

    def __init__(self):
        super().__init__()
        self.model_name = "random"

    def _get_sorted_exchange_actor_list(self):
        all_exchange_actors = []
        for exchange in self.exchanges:  # type: RandomRateExchange

            if exchange.is_valid:
                all_exchange_actors.append(exchange.i)
                all_exchange_actors.append(exchange.j)

        all_exchange_actors.sort(key=attrgetter("eu"), reverse=True)

        return all_exchange_actors

    def sort_exchanges(self):
        pass

    def highest_gain(self):

        deadlock = defaultdict(int)

        while len(self.exchanges) > 0:

            exchange_actors = self._get_sorted_exchange_actor_list()

            # place each ExchangeActor in a dict where al his exchanges are ordered by the utility
            exchange_actors_by_gain = defaultdict(lambda: defaultdict(list))

            # sort all the gains by actor in a dict
            # so we can sort this DESC
            # now its possible to check which gain is the next in line to be lowered.
            # this is always the second element, because the first element is the highest gain.
            exchange_actors_by_actor = defaultdict(list)

            highest_actors = {}
            exchange_by_key = {}

            deadlock[len(self.exchanges)] += 1

            for exchange_actor in exchange_actors:  # type: RandomRateExchangeActor

                exchange_actors_by_actor[exchange_actor.actor.name].append(
                    exchange_actor.eu
                )
                exchange_actors_by_gain[exchange_actor.actor.name][
                    exchange_actor.eu
                ].append(exchange_actor)

            if deadlock[len(self.exchanges)] > 1024:
                # this should never happen?
                logging.info("deadlock")

                for ex in self.exchanges:
                    logging.info(ex)

                self.exchanges.clear()
                return None

            # is there an exchange where both actors archive their highest gain
            for exchange_actor in exchange_actors:
                # if an actor is not yet in the list, add him
                # the list is sorted by gain so the first result is the exchange with the highest gain
                # elif, if the actor experience a gain that is equal to the previous highest gain, mark the exchange
                # to the highest also, return the exchange when this already is the case.
                if exchange_actor.actor_name not in highest_actors:
                    highest_actors[exchange_actor.actor_name] = exchange_actor

                    # if the exchange is not yet present, mark this exchange as an exchange where on of the both actors
                    # achieves his highest gain
                    # otherwise this exchange is already marked, then we have found an exchange where both actors
                    # achieve there highest gain
                    if exchange_actor.exchange.key not in exchange_by_key:
                        exchange_by_key[
                            exchange_actor.exchange.key
                        ] = exchange_actor.exchange
                    else:
                        self.remove_exchange_by_key(exchange_actor.exchange.key)
                        return exchange_actor.exchange

                elif highest_actors[exchange_actor.actor_name].eu == exchange_actor.eu:
                    if exchange_actor.exchange.key not in exchange_by_key:
                        exchange_by_key[
                            exchange_actor.exchange.key
                        ] = exchange_actor.exchange
                    else:
                        self.remove_exchange_by_key(exchange_actor.exchange.key)
                        return exchange_actor.exchange

            # lower highest gains
            for actor_name, values in exchange_actors_by_actor.items():
                # do nothing when the actor has one exchange left because this exchange is automatically his highest
                if len(values) > 1:
                    highest = values[0]

                    second_highest = self._find_first_element_not_equal(values)
                    # if all the exchanges have the same gain, do nothing
                    if second_highest is not None:

                        delta_eu = highest - second_highest

                        highest_exchanges = exchange_actors_by_gain[actor_name][highest]
                        # lower ALL the exchanges with the highest gain
                        for highest_exchange_actor in highest_exchanges:

                            highest_exchange = (
                                highest_exchange_actor.exchange
                            )  # type: RandomRateExchange
                            opposite_actor_name = (
                                highest_exchange_actor.opposite_actor.actor_name
                            )

                            if (
                                highest_exchange[actor_name].y
                                == highest_exchange[opposite_actor_name].x_demand
                            ):
                                highest_exchange[opposite_actor_name].recalculate(
                                    delta_eu, increase=False
                                )
                            elif (
                                highest_exchange[opposite_actor_name].y
                                == highest_exchange[actor_name].x_demand
                            ):
                                highest_exchange[actor_name].recalculate(
                                    delta_eu, increase=True
                                )
                            else:
                                highest_exchange[opposite_actor_name].recalculate(
                                    delta_eu, increase=False
                                )

                            highest_exchange[opposite_actor_name].adjust_utility(
                                delta_eu
                            )

    @staticmethod
    def new_exchange_factory(i, j, p, q, model, groups):
        """
        Creates a new instance of the RandomRateExchange
        """
        return RandomRateExchange(i, j, p, q, model, groups)

    def remove_exchange_by_key(self, key):

        for __, exchange in enumerate(self.exchanges):

            if key == exchange.key:
                del self.exchanges[__]
                return

    @staticmethod
    def _find_first_element_not_equal(exchange_utility_list: List[Decimal]):

        previous_value = exchange_utility_list.pop(0)
        for value in exchange_utility_list:
            if abs(value - previous_value) > 0:
                return value



import logging
from collections import defaultdict
from decimal import Decimal
from itertools import combinations


class Issue:
    def __init__(self, name, lower=None, upper=None):
        """
        Refers an issue
        :param name: str
        :param lower: int
        :param upper: int
        """
        self.delta = 0
        self.step_size = 0

        self.name = name

        self.lower = lower
        self.upper = upper

        self.comment = ""

        if self.lower is not None or self.upper is not None:
            self.calculate_delta()
            self.calculate_step_size()

    @property
    def issue_id(self):
        return self.name

    def calculate_delta(self):
        self.delta = self.upper - self.lower

    def calculate_step_size(self):
        if self.delta is not 0:
            self.step_size = Decimal(100 / self.delta)
        else:
            self.step_size = 0

    def de_normalize(self, value):
        if value == 0:
            return self.lower

        return value / self.step_size + self.lower

    def normalize(self, value):
        return Decimal(value - self.lower) * self.step_size

    def expand_lower(self, value):

        if self.lower is None:
            self.lower = value
        elif value < self.lower:
            self.lower = value

    def expand_upper(self, value):
        if self.upper is None:
            self.upper = value
        elif value > self.upper:
            self.upper = value

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "{0} [{1} - {2}]".format(self.name, self.lower, self.upper)

    def __eq__(self, other):

        if isinstance(other, str):
            b = self.issue_id == other
            return b

        if isinstance(other, Issue):
            return self.issue_id == other.issue_id

        if isinstance(other, int):
            h = self.__hash__()
            b = h == other
            return b

        if not other:
            return False

        raise NotImplementedError()

    def __hash__(self):
        return hash(self.issue_id)

    def __lt__(self, other):
        """
        Needed for sorting
        :param other:
        :return:
        """
        return self.issue_id < other.issue_id


class Actor:
    def __init__(self, name, actor_id=None):
        """
        Represents an Actor
        """
        self.name = name
        self.comment = ""

        if actor_id:
            self.actor_id = actor_id
        else:
            self.actor_id = name

    def __eq__(self, other):

        from decide.data.database import Actor as ModelActor

        if isinstance(other, str):
            return self.actor_id == str(other)
        if isinstance(other, int):
            return self.__hash__() == other
        if isinstance(other, Actor):
            return self.actor_id == other.actor_id
        if isinstance(other, ModelActor):
            return self.actor_id == other.key
        if not other:
            return False
        raise NotImplementedError()

    def __lt__(self, other):
        """
        Needed for sorting
        :param other:
        :return:
        """
        if not isinstance(other, Actor):
            raise ValueError("error")

        return self.actor_id < other.actor_id

    def __hash__(self):
        """
        Hashing is based on the id
        :return:
        """
        return hash(self.actor_id)

    def __str__(self):
        """
        Human representation of this object
        :return:
        """
        return self.name


class ActorIssue:
    """
    Represents a combination between an actor and issue
    """

    def __init__(self, actor: Actor, issue: Issue, position, salience, power):
        """
        :param actor: Actor 
        :param issue: Issue
        :param position: Double
        :param salience: Double
        :param power: Double
        """

        self.actor = actor
        self.power = Decimal(power)
        self.position = Decimal(position)
        self.salience = Decimal(salience)
        self.left = False  # left of nbs
        self.issue = issue

    def is_left_to_nbs(self, nbs):
        """
        True when the position is lower than the Nash Bargaining Solution for this issue. 
        :param nbs: int
        :return: bool 
        """
        self.left = self.position < nbs
        return self.left

    def __str__(self):
        return "{0} on {1} with x={2}, s={3}, c={4}".format(
            self.actor.name, self.issue.name, self.position, self.salience, self.power
        )

    def __eq__(self, other):
        """
        True when the name and issue are the same
        :param other: 
        :return: 
        """
        if isinstance(other, ActorIssue):
            return self.actor == other.actor and self.issue == other.issue

        return NotImplemented

    def __lt__(self, other):
        return self.actor < other.actor


class DemandActorIssue(ActorIssue):
    """
    Object for a demand issue, has the same properties as a ActorIssue
    """

    def __init__(self, actor_issue: ActorIssue):
        super().__init__(
            actor_issue.actor,
            actor_issue.issue,
            actor_issue.position,
            actor_issue.salience,
            actor_issue.power,
        )

        self.actor_issue = actor_issue


class SupplyActorIssue(ActorIssue):
    """
    Object for a demand issue, has a extra voting_position (self.y)
    """

    def __init__(self, actor_issue: ActorIssue, y=None):
        super().__init__(
            actor_issue.actor,
            actor_issue.issue,
            actor_issue.position,
            actor_issue.salience,
            actor_issue.power,
        )

        self.y = y
        self.actor_issue = actor_issue


class AbstractExchangeActor:
    """
    Represents an exchange actor. Contains his demand and supply issues and voting-position
    """

    def __init__(
            self,
            model: "AbstractModel",
            actor: Actor,
            demand_issue: Issue,
            supply_issue: Issue,
            exchange: "AbstractExchange",
    ):
        """
        Constructor, must be invoked

        :param model: AbstractModel
        :param actor: Actor
        :param demand_issue: Issue
        :param supply_issue: Issue
        :param exchange: AbstractExchange
        """
        # the actor
        self.actor = actor

        # create supply and demand issues from the ActorIssues
        self.supply = SupplyActorIssue(model.actor_issues[supply_issue][actor])
        self.demand = DemandActorIssue(model.actor_issues[demand_issue][actor])

        # the voting position for the supply issue
        self.y = None
        self.start_position = self.supply.position

        # the expected utility
        self.eu = 0

        self.opposite_actor = None  # type: AbstractExchangeActor
        # current move (should be equal with self.y - self.supply.position
        self.move = 0

        # history of moves
        self.moves = []

        # placeholder for the nbs before this exchange and after
        self.nbs_0 = 0
        self.nbs_1 = 0

        # backref to the exchange object and model
        self.exchange = exchange
        self.model = model

        self.is_adjusted_by_nbs = False

    def is_move_valid(self, move):

        """
        Cheks if a move not exceeds the interval [0,100]

        :param move:
        :return:
        """
        if abs(move) > 100 or abs(move) <= 1e-10:
            return False

        # if an exchange is on the edges there is no move posible
        if self.supply.position + move < 0 or self.supply.position + move > 100:
            return False

        if sum(self.moves) > 100:
            return False

        moves_min = min(self.moves)
        moves_max = max(self.moves)

        if (moves_min < 0 and moves_max < 0) or (moves_min > 0 and moves_max > 0):
            return True

    def new_start_position(self):
        """
        Calculate the new starting postion for the next round
        :return:
        """
        from decide.model import calculations

        return calculations.new_start_position(
            salience=self.supply.salience,
            x=self.start_position,
            y=self.y,
            salience_weight=self.model.SALIENCE_WEIGHT,
            fixed_weight=self.model.FIXED_WEIGHT,
        )

    def actor_issues(self):
        """ shortcut, demand ActorIssues should never not be needed """
        return self.supply_actor_issues()

    def supply_actor_issues(self):
        """ shortcut function"""
        return self.model.actor_issues[self.supply.issue]

    def demand_actor_issues(self):
        """ shortcut function"""
        return self.model.actor_issues[self.demand.issue]

    def adjust_nbs(self, position):

        actor_issues = self.actor_issues()

        updates = self.exchange.updates[self.supply.issue]

        from . import calculations  # Todo should be global import

        return calculations.adjusted_nbs(
            actor_issues=actor_issues,
            updates=updates,
            actor=self.actor,
            new_position=position,
            denominator=self.model.nbs_denominators[self.supply.issue],
        )

    def equals_actor_demand_issue(self, other: "AbstractExchangeActor"):
        """
        Has this ExchangeActor the same actor and same demand issue
        :param other:
        :return:
        """
        return self.equals_actor(other) and self.equals_demand_issue(other)

    def equals_actor_supply_issue(self, other: "AbstractExchangeActor"):
        """
        Has this ExchangeActor the same actor and same supply issue
        :param other:
        :return:
        """
        return self.equals_actor(other) and self.equals_supply_issue(other)

    def equals_actor(self, other):
        """
        Has this ExchangeActor the same actor
        :param other:
        :return:
        """
        return self.actor == other.actor

    def equals_supply_issue(self, other: "AbstractExchangeActor"):
        """
        Has this ExchangeActor the same supply issue
        :param other: AbstractExchangeActor
        :return:
        """
        return self.supply.issue == other.supply.issue

    def equals_demand_issue(self, other: "AbstractExchangeActor"):
        """
        Has this ExchangeActor the same demand issue?
        :param other: AbstractExchangeActor
        :return:
        """
        return self.demand.issue == other.demand.issue

    def check_nbs(self):
        """
        Calculate if the outcome doesn't shifts over the original position of the demand actor of his first exchange.
        :return:
        """

        from . import calculations

        self.nbs_0 = self.adjust_nbs(self.supply.position)

        self.nbs_1 = self.adjust_nbs(self.y)

        x = self.opposite_actor.demand.position

        check = round(self.opposite_actor.demand.position, 3) >= round(
            self.nbs_0, 3
        ) and round(self.opposite_actor.demand.position, 3) >= round(self.nbs_1, 3)
        check_2 = round(self.opposite_actor.demand.position, 3) <= round(
            self.nbs_0, 3
        ) and round(self.opposite_actor.demand.position, 3) <= round(self.nbs_1, 3)

        if (x - self.nbs_0) >= 0 and (x - self.nbs_1) >= 0:
            if not check:
                raise Exception("test")
        elif (x - self.nbs_0) <= 0 and (x - self.nbs_1) <= 0:
            if not check_2:
                raise Exception("test123")
        else:
            delta = abs(
                calculations.adjusted_nbs_by_position(
                    actor_issues=self.actor_issues(),
                    updates=self.exchange.updates[self.supply.issue],
                    actor=self.actor,
                    x_pos=self.supply.position,
                    new_nbs=self.opposite_actor.demand.position,
                    denominator=self.model.nbs_denominators[self.supply.issue],
                )
            )

            self.exchange.dp = calculations.exchange_ratio(
                delta_x=delta,
                salience=self.supply.salience,
                power=self.supply.power,
                denominator=self.model.nbs_denominators[self.supply.issue],
            )

            self.exchange.dq = calculations.by_exchange_ratio(self, self.exchange.dp)

            self.opposite_actor.move = abs(
                calculations.reverse_move(
                    actor_issues=self.opposite_actor.actor_issues(),
                    actor=self.opposite_actor,
                    exchange_ratio=self.exchange.dq,
                )
            )

            self.move = delta

            if self.opposite_actor.supply.position > self.demand.position:
                self.opposite_actor.move *= -1

            if self.supply.position > self.opposite_actor.demand.position:
                self.move *= -1

            self.opposite_actor.moves.pop()
            self.moves.pop()
            self.opposite_actor.moves.append(self.opposite_actor.move)
            self.moves.append(self.move)

            self.opposite_actor.y = (
                    self.opposite_actor.supply.position + self.opposite_actor.move
            )
            self.y = self.supply.position + self.move

            eui = calculations.expected_utility(
                self.opposite_actor, self.exchange.dq, self.exchange.dp
            )
            euj = calculations.expected_utility(
                self, self.exchange.dp, self.exchange.dq
            )

            if abs(eui - euj) > 0.0001:
                raise Exception("Expected equal gain")
            else:
                self.exchange.gain = abs(eui)

            b1 = self.opposite_actor.is_move_valid(self.opposite_actor.move)
            b2 = self.is_move_valid(self.move)

            self.nbs_1 = self.adjust_nbs(self.y)

            if abs(self.nbs_1 - self.y) > 1e-10:
                logging.debug(
                    "The new calculated NBS is not exactly the same as the new position of the actor (nbs 0 {} nbs 1: {} y: {})".format(
                        self.nbs_0, self.nbs_1, self.y
                    )
                )

            self.exchange.is_valid = b1 and b2

            self.is_adjusted_by_nbs = True

    def __str__(self):
        """
        The String representation of the object
        :return: String representation of the object
        """

        if hasattr(self, "eu"):
            return "{0} {1} eu={2}, x={3} y={4}, d={5}".format(
                self.actor.name,
                self.supply.issue.name,
                round(self.eu, 2),
                self.supply.position,
                self.y,
                self.opposite_actor.demand.position,
            )
        else:
            return "{0} {1} {2} {3} ({4})".format(
                self.actor.name,
                self.supply.issue.name,
                self.supply.position,
                self.y,
                self.opposite_actor.demand.position,
            )

    def __eq__(self, other):
        return (
                self.equals_actor(other)
                and self.equals_demand_issue(other)
                and self.equals_supply_issue(other)
        )


class AbstractExchange:
    """
    An exchange between two actors and two issues. Each actor has a demand and supply issue
    """

    actor_class = AbstractExchangeActor

    def __init__(self, i, j, p, q, m, groups):
        """
        An exchange between two actors and two issues. Each actor has a demand and supply issue
        :param i: Actor
        :param j: Actor
        :param p: Issue
        :param q: Issue
        :param m: AbstractModel
        :param groups: tuple of the two groups each actor is in
        """
        self.model = m
        self.groups = groups

        # if the groups are not valid exit
        self.validate_groups()

        self.gain = 0  # expected utility
        self.is_valid = True
        self.re_calc = False
        self.p = p
        self.q = q

        self.dp = 0
        self.dq = 0

        self.updates = defaultdict(dict)

        # c.	If (1) holds, i shifts his position on issue p in the direction of j,
        # whereas j shifts his position on issue q in the direction of i.
        # Issue p is then called the supply issue of i and the demand issue of j,
        # whereas issue q is the demand issue of i and the supply issue of j.
        # If (2) holds,
        # issue q is the supply issue of i and issue p is the supply issue of j.
        # if ( (model$s_matrix[p, i] / model$s_matrix[q, i]) < (model$s_matrix[p, j] / model$s_matrix[q, j]))
        if (m.get_value(i, p, "s") / m.get_value(i, q, "s")) < (
                m.get_value(j, p, "s") / m.get_value(j, q, "s")
        ):
            self.i = self.actor_class(
                m, j, supply_issue=q, demand_issue=p, exchange=self
            )
            self.j = self.actor_class(
                m, i, supply_issue=p, demand_issue=q, exchange=self
            )
        else:
            self.i = self.actor_class(
                m, i, supply_issue=q, demand_issue=p, exchange=self
            )
            self.j = self.actor_class(
                m, j, supply_issue=p, demand_issue=q, exchange=self
            )

        self.j.opposite_actor = self.i
        self.i.opposite_actor = self.j

    def calculate(self):
        """Method stub to be overriden"""
        raise NotImplementedError

    def invalidate_move(self):
        self.i.moves.pop()
        self.j.moves.pop()

    def invalidate_exchange_by_supply(self, exchange):

        """
        If the actors and supply issues match one or both the actors from the exchange that is executed,
        this exchange invalidates and needs to be recalculated.

        When the move invalidates, erase the current values for the move so the new values can be calculated.

        If both the cases happen (i equals j and j equals i) the exchange is identical, this is not posible.

        :param exchange:
        :return:
        """
        invalid_i = False
        invalid_j = False

        if self.i.equals_actor_supply_issue(exchange.i):
            self.i.supply.position = exchange.i.y
            self.i.moves.append(exchange.i.moves[-1])
            invalid_i = True
        elif self.i.equals_actor_supply_issue(exchange.j):
            self.i.supply.position = exchange.j.y
            self.i.moves.append(exchange.j.moves[-1])
            invalid_i = True

        if self.j.equals_actor_supply_issue(exchange.i):
            self.j.supply.position = exchange.i.y
            self.j.moves.append(exchange.i.moves[-1])
            invalid_j = True
        elif self.j.equals_actor_supply_issue(exchange.j):
            self.j.supply.position = exchange.j.y
            self.j.moves.append(exchange.j.moves[-1])
            invalid_j = True

        if invalid_i or invalid_j:
            self.invalidate_move()
        elif invalid_i and invalid_j:
            raise Exception(
                "Exchanges are equal: {0} and {1}".format(str(self), str(exchange))
            )

        return invalid_i or invalid_j

    def update_updates(
            self,
            exchange_actor: AbstractExchangeActor,
            demand_issue: Issue,
            updated_position,
    ):
        """
        Update the updates dictionary with the current exchange
        :param exchange_actor:
        :param demand_issue:
        :param updated_position:
        """

        # if the start position is left of the updated_position then TODO
        if exchange_actor.start_position <= updated_position:
            # what is this?
            if updated_position < exchange_actor.y:
                self.updates[demand_issue][exchange_actor.actor] = updated_position
            else:
                self.updates[demand_issue][exchange_actor.actor] = exchange_actor.y
        else:
            if updated_position > exchange_actor.y:
                self.updates[demand_issue][exchange_actor.actor] = updated_position
            else:
                self.updates[demand_issue][exchange_actor.actor] = exchange_actor.y

    def recalculate(self, exchange):
        """
        This exchange needs to be recalculated because the given exchange is performed
        and has an influence on this (self) exchange
        :param exchange:
        """
        self.re_calc = self.invalidate_exchange_by_supply(exchange)

        # update the positions for the demand actors...
        if exchange.j.equals_actor_demand_issue(
                self.j
        ) or exchange.j.equals_actor_demand_issue(self.i):

            if exchange.i.actor in self.updates[exchange.j.demand.issue]:
                self.update_updates(
                    exchange.i,
                    exchange.j.demand.issue,
                    self.updates[exchange.j.demand.issue][exchange.i.actor],
                )
            else:
                self.updates[exchange.j.demand.issue][exchange.i.actor] = exchange.i.y

            if not self.re_calc:
                self.invalidate_move()
                self.re_calc = True

        if exchange.i.equals_actor_demand_issue(
                self.j
        ) or exchange.i.equals_actor_demand_issue(self.i):

            if exchange.j.actor in self.updates[exchange.i.demand.issue]:
                self.update_updates(
                    exchange.j,
                    exchange.i.demand.issue,
                    self.updates[exchange.i.demand.issue][exchange.j.actor],
                )
            else:
                self.updates[exchange.i.demand.issue][exchange.j.actor] = exchange.j.y

            if not self.re_calc:
                self.invalidate_move()
                self.re_calc = True

        if self.re_calc:
            self.calculate()

    def validate_groups(self):
        """
        Two actors can only exchange when they are in group a-d or b-c
        :return:
        """
        if "a" in self.groups and "d" in self.groups:
            return True
        elif "b" in self.groups and "c" in self.groups:
            return True
        else:
            raise Exception(
                "invalid group combination [%,%]".format(self.groups[0], self.groups[1])
            )

    def get_inner_groups(self):
        """
        When groups[0] is not a and not d, the b&c group is inner.
        If groups[0] == a, then groups[1] has to be D
        """

        if len(self.groups) != 2:
            raise ValueError("Too much groups given!")

        if "a" in self.groups and "d" in self.groups:
            return ["a", "d"]

        if "b" in self.groups and "c" in self.groups:
            return ["b", "c"]

        raise ValueError(
            "Actors should only exchange with their inner group."
            " The given groups are {0}".format(self.groups)
        )

    def __str__(self):
        return "{0}: {1}, {2}".format(round(self.gain, 9), str(self.i), str(self.j))

    def __bool__(self):
        return self.is_valid

    def __eq__(self, other):
        return self.i == other.i and self.j == other.j


class AbstractModel:
    SALIENCE_WEIGHT = 0.4
    FIXED_WEIGHT = 0.1
    VERBOSE = True  # verbose messages for debugging

    def __init__(self, *args, **kwargs):
        self.issues = {}
        self.actor_issues = defaultdict(dict)
        self.actors = {}
        self.exchanges = []
        self.nbs = {}
        self.issue_combinations = []
        self.groups = {}
        self.moves = {}  # dict with issue,actor[move_1,move_2,move_3]
        self.nbs_denominators = {}
        self.data_set_name = ""
        self.model_name = "abstract"
        self.tie_count = 0

    def get_actor_issue(self, actor: Actor, issue: Issue):
        """
        Getter function for an ActorIssue
        :param actor: the id of the actor
        :param issue: the id of the issue
        :return:
        """
        if actor in self.actor_issues[issue]:
            return self.actor_issues[issue][actor]
        else:
            return False

    def get_value(self, actor: Actor, issue: Issue, field):
        """
        Get the value of an attribute for an actor on the specified issue
        :param actor: str
        :param issue: str
        :param field: str
        :return: Double
        """

        a = self.actor_issues[issue][actor]

        if a is not False:

            if field is "c":
                return a.power
            if field is "s":
                return a.salience
            if field is "x":
                return a.position

        raise ValueError("ActorIssue not found")

    def add_actor(self, actor_name, actor_id=None, comment="") -> Actor:
        """
        Add an actor to the model
        :param comment:
        :param actor_id:
        :param actor_name:
        :return:
        """
        actor = Actor(actor_name, actor_id)
        actor.comment = comment
        self.actors[actor] = actor
        return actor

    def add_issue(self, issue_name, issue_id=None, comment="") -> Issue:
        """
        Add an issue to the model
        :param comment:
        :param issue_id:
        :param issue_name:
        """
        issue = Issue(issue_name)
        issue.comment = comment
        self.issues[issue] = issue
        return issue

    def add_actor_issue(self, actor, issue, position, salience, power):
        """
        Add an actor issue to the model
        :param actor: Actor
        :param issue: Issue
        :param position: Double
        :param salience: Double
        :param power: Double
        :return:
        """

        if not isinstance(actor, Actor):
            if actor in self.actors:
                actor = self.actors[actor]
            elif hash(actor) in self.actors:
                actor = self.actors[hash(actor)]
            else:
                raise ValueError("Actor not found: {0}".format(actor))

        if not isinstance(issue, Issue):
            if issue in self.issues:
                issue = self.issues[issue]
            elif hash(issue) in self.issues:
                issue = self.issues[hash(issue)]
            else:
                raise ValueError(
                    "The issue '{0}' is missing in your input file. "
                    "Define '#P' with '{0}' and add a description".format(issue)
                )

        self.actor_issues[issue][actor] = ActorIssue(
            actor, issue, position, salience, power
        )

        return self.actor_issues[issue][actor]

    def add_exchange(self, i, j, p, q, groups):
        """
        Add an exchange pair to the model
        :param i:
        :param j:
        :param p:
        :param q:
        :param groups:
        :return:
        """
        e = self.new_exchange_factory(i, j, p, q, self, groups)
        e.calculate()
        self.exchanges.append(e)
        return e

    def calc_nbs(self):
        from . import calculations

        """
        Calculate the nash bargaining solution for all the issue
        """
        for issue, actor_issues in self.actor_issues.items():
            self.nbs_denominators[issue] = calculations.calc_nbs_denominator(
                actor_issues
            )

            nbs = calculations.nash_bargaining_solution(
                actor_issues, self.nbs_denominators[issue]
            )
            self.nbs[issue] = nbs

    def determine_positions(self):
        """
        Determine if the position of an actor is left or right of the Nash Bargaining Solution on an issue
        """
        for issue_name, issue_nbs in self.nbs.items():
            for actor_issue in self.actor_issues[issue_name].values():
                actor_issue.is_left_to_nbs(issue_nbs)

    def calc_combinations(self):
        """
        Create a list of all possible combinations for the issues
        """
        self.issue_combinations = combinations(self.issues, 2)

    def determine_groups(self):
        """
        There are 4 groups: A, B, C, and D.
        An actor is member of group A if his position on both issues is left of the NBS.
        Each actor of group A can exchange with the actors of Group D, the actors of B with C.
        """
        for combination in self.issue_combinations:

            pos = [[], [], [], []]

            for k, actor in self.actors.items():

                a0 = self.get_actor_issue(actor=actor, issue=combination[0])
                a1 = self.get_actor_issue(actor=actor, issue=combination[1])

                # some magic happens here: we have four possibilities and two bytes, so
                # A = 00 = 0
                # B = 01 = 1
                # C = 10 = 2
                # D = 11 = 3

                if a0 is not False and a1 is not False:
                    position = a0.left | a1.left * 2
                    pos[position].append(actor)

            combination_id = "{0}-{1}".format(combination[0], combination[1])

            self.groups[combination_id] = {
                "a": pos[0],
                "b": pos[1],
                "c": pos[2],
                "d": pos[3],
            }

            # all actors of group A and D
            for i in pos[0]:
                for j in pos[3]:
                    self.add_exchange(
                        i, j, combination[0], combination[1], groups=["a", "d"]
                    )

                    self.actor_issues[combination[0]][i].group = "a"
                    self.actor_issues[combination[1]][j].group = "d"

            # all actors of group B and C
            for i in pos[1]:
                for j in pos[2]:
                    self.add_exchange(
                        i, j, combination[0], combination[1], groups=["b", "c"]
                    )
                    self.actor_issues[combination[0]][i].group = "b"
                    self.actor_issues[combination[1]][j].group = "c"

    def remove_invalid_exchanges(self, res):
        """
        Removes all the invalid exchanges from the exchanges list and return them
        :param res: a list of exchanges
        :return: a list with online valid exchanges
        """
        length = len(self.exchanges)

        valid_exchanges = []
        invalid_exchanges = []

        for i in range(length):

            if self.exchanges[i].is_valid:
                self.exchanges[i].recalculate(res)

                if self.exchanges[i].is_valid:
                    valid_exchanges.append(self.exchanges[i])
                else:
                    invalid_exchanges.append(self.exchanges[i])

        self.exchanges = valid_exchanges

        return invalid_exchanges

    def sort_exchanges(self):
        """
        Abstract method
        :return:
        """
        raise NotImplementedError

    def highest_gain(self):
        """
        Abstract method
        :return:
        """
        raise NotImplementedError

    def create_existing_issue_set_key(self, p, q):
        """
        Create a combination of issue so it can be used as key
        """
        issue_set_key = "{0}-{1}".format(p, q)
        # an combination only exists once, so it can happen that we have to change the sequence of the keys
        if issue_set_key not in self.groups:
            issue_set_key = "{0}-{1}".format(q, p)

        return issue_set_key

    def is_inner_group_member(self, actor, inner, issue_set_key):
        """
        When an actor is in the same [A] or opposing group [D], return true
        False, When in [B] or [C]
        :param actor:
        :param issue_set_key:
        :param inner:
        :return:
        """

        is_inner = (
                actor in self.groups[issue_set_key][inner[0]]
                or actor in self.groups[issue_set_key][inner[1]]
        )

        return is_inner

    @staticmethod
    def new_exchange_factory(i, j, p, q, model_ref, groups):
        """
        Creates a new Exchange set 
        :param groups: 
        :param model_ref: 
        :param q: 
        :param p: 
        :param j: 
        :param i:
        """
        raise NotImplementedError



from decide.qt.mainwindow.gui import DecideMainWindow
from decide.qt.mainwindow.settings import SettingsFormWidget, ProgramSettings


def test_load_settings(qtbot):
    """
    Load settings from ProgramSettings
    """

    program_settings = ProgramSettings()

    main_window = DecideMainWindow()

    salience_weight = 0.99

    program_settings.salience_weight = salience_weight

    assert program_settings.salience_weight == salience_weight

    widget = SettingsFormWidget(program_settings, main_window)
    widget.load()

    assert widget.salience_weight.value() == salience_weight

# def test_save_settings(qtbot):
#     """
#     Can we save settings from the widget to ProgramSettings
#     """
#
#     settings = SettingsFormWidget(None, None)
#     settings.load()
#
#     pytest.fail()



import os

import pandas as pd
from decide import data_folder
from decide.data.database import connection, Manager
from decide.results.helpers import list_to_sql_param

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)


def write_summary_result(conn, model_run_ids, output_directory):
    sql = """SELECT SUM(e.own)            as own,
           SUM(e.inner_positive) as inner_positive,
           SUM(e.inner_negative) as inner_negative,
           SUM(e.outer_positive) as outer_positive,
           SUM(e.outer_negative) as outer_negative,
           a.name                   actor,
           m.p
    FROM externality e
             JOIN actor a on e.actor_id = a.id
             JOIN iteration i on e.iteration_id = i.id
             JOIN repetition r on i.repetition_id = r.id
             JOIN modelrun m on r.model_run_id = m.id         
    WHERE m.id IN(%s)
    GROUP BY e.actor_id, m.id""" % list_to_sql_param(model_run_ids)

    df = pd.read_sql_query(
        sql,
        con=conn,
        index_col=['p'],
    )

    df_inner_positive = pd.pivot_table(df, index=['p'], columns=['actor'], values=['inner_positive'])
    df_inner_positive.to_csv(os.path.join(output_directory, 'externalities_inner_positive.csv'))

    plt = df_inner_positive.plot()

    lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.set_title('Inner Positive')
    plt.set_xticks(df_inner_positive.index)
    plt.figure.savefig(os.path.join(
        output_directory, 'externalities_inner_positive.png'),
        bbox_extra_artists=(lgd,),
        bbox_inches="tight",
    )

    df_inner_negative = pd.pivot_table(df, index=['p'], columns=['actor'], values=['inner_negative'])
    df_inner_negative.to_csv(os.path.join(output_directory, 'externalities_inner_negative.csv'))

    plt = df_inner_negative.plot()

    lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.set_title('Inner Negative')
    plt.set_xticks(df_inner_negative.index)
    plt.figure.savefig(os.path.join(
        output_directory, 'externalities_inner_negative.png'),
        bbox_extra_artists=(lgd,),
        bbox_inches="tight",
    )

    # OUTER POSITIVE

    df_outer_positive = pd.pivot_table(df, index=['p'], columns=['actor'], values=['outer_positive'])
    df_outer_positive.to_csv(os.path.join(output_directory, 'externalities_outer_positive.csv'))

    plt = df_outer_positive.plot()

    lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.set_title('Outer Positive')
    plt.set_xticks(df_outer_positive.index)
    plt.figure.savefig(os.path.join(
        output_directory, 'externalities_outer_positive.png'),
        bbox_extra_artists=(lgd,),
        bbox_inches="tight",
    )

    # OUTER NEGATIVE

    df_outer_negative = pd.pivot_table(df, index=['p'], columns=['actor'], values=['outer_negative'])
    df_outer_negative.to_csv(os.path.join(output_directory, 'externalities_outer_negative.csv'))

    plt = df_outer_negative.plot()

    lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.set_title('Outer negative')
    plt.set_xticks(df_outer_negative.index)
    plt.figure.savefig(
        os.path.join(output_directory, 'externalities_outer_negative.png'),
        bbox_extra_artists=(lgd,),
        bbox_inches="tight",
    )


if __name__ == '__main__':
    from environs import Env

    env = Env()
    Env.read_env()  # read .env file, if it exists

    url = env('DATABASE_URL')

    m = Manager(url)
    m.init_database()

    model_run_id = 1

    write_summary_result(connection, [model_run_id], data_folder)



from PyQt5 import QtCore

from decide.qt.inputwindow.gui import InputWindow, register_app


def test_actor_widget_add_button(qtbot):
    x = InputWindow()
    register_app(x)

    assert x.actor_widget._row_pointer == 1, 'Only the heading is shown'
    assert x.actor_widget.grid_layout.rowCount() == 1

    qtbot.mouseClick(x.actor_widget.add_button, QtCore.Qt.LeftButton)

    assert x.actor_widget.grid_layout.rowCount() == 2, 'The heading and an empty row is shown'

# FIX killing the window
# def test_comment_button(qtbot):
#     x = ActorWidget()
#
#     x.add_actor('PyTest', 1)
#
#     assert x._row_pointer == 2, "Only the heading and an actor"
#
#     comment_button = x.grid_layout.itemAtPosition(1, 2).widget()
#
#     qtbot.mouseClick(comment_button, QtCore.Qt.LeftButton)
#
#     assert comment_button
#
#
# def test_load_comment(qtbot):
#     x = InputWindow()
#     register_app(x)
#
#     x.load_sample_data()
#
#     comment_button = x.actor_widget.grid_layout.itemAtPosition(1, 2).widget()
#
#     qtbot.mouseClick(comment_button, QtCore.Qt.LeftButton)
#
#     assert True, 'Will fail on the mouse click'



from decide.data import types, reader
from decide.data.modelfactory import ModelFactory
from decide.model.equalgain import EqualGainModel


def test_issue_filter():
    data_file = reader.InputDataFile()
    data_file.parse_rows([
        [types.PartialIssue.starts_with, 'Issue #1', 'description'],
        [types.PartialIssue.starts_with, 'Issue #2', 'description'],
        [types.PartialIssue.starts_with, 'Issue #3', 'description'],
    ])

    factory = ModelFactory(data_file, issue_whitelist=['Issue #2', 'Issue #3'])

    assert len(factory.filter_issues()) == 2
    assert 'Issue #1' not in factory.filter_issues()


def test_actor_filter():
    data_file = reader.InputDataFile()
    data_file.parse_rows([
        [types.PartialActor.starts_with, 'Actor #1', 'description'],
        [types.PartialActor.starts_with, 'Actor #2', 'description'],
        [types.PartialActor.starts_with, 'Actor #3', 'description'],
    ])

    factory = ModelFactory(data_file, actor_whitelist=['Actor #2', 'Actor #3'])

    assert len(factory.filter_actors()) == 2
    assert 'Actor #1' not in factory.filter_actors()


def test_actor_issue_filter():
    actor_1 = 'Actor #1'
    actor_2 = 'Actor #2'
    actor_3 = 'Actor #3'

    actors = [actor_1, actor_2, actor_3]
    actor_whitelist = actors[1:]

    issue_1 = 'Issue #1'
    issue_2 = 'Issue #2'
    issue_3 = 'Issue #3'

    issues = [issue_1, issue_2, issue_3]
    issue_whitelist = issues[1:]

    data = []

    for issue in issues:
        data.append([types.PartialIssue.starts_with, issue, issue])

    for actor in actors:
        data.append([types.PartialActor.starts_with, actor, actor])

        for issue in issues:
            data.append([types.ActorIssue.starts_with, actor, issue, 100, 1, 1])

    data_file = reader.InputDataFile()
    data_file.parse_rows(data)
    data_file.update_issues_with_positions()

    assert len(data_file.errors) == 0
    assert len(data_file.actors) == 3
    assert len(data_file.issues) == 3
    assert len(data_file.actor_issues) == 9

    factory = ModelFactory(
        date_file=data_file,
        actor_whitelist=actor_whitelist,
        issue_whitelist=issue_whitelist
    )

    filtered_issues = factory.filter_issues()
    filtered_actors = factory.filter_actors()

    assert len(filtered_issues) == 2
    assert len(filtered_actors) == 2

    filtered_actor_issues = factory.filter_actor_issues(filtered_actors, filtered_issues)

    assert len(filtered_actor_issues) == 4

    model = factory(EqualGainModel)

    assert len(model.issues) == len(filtered_issues)
    assert len(model.actors) == len(filtered_actors)
    assert len(model.actor_issues) == 2

    count = sum(len(x) for x in model.actor_issues.values())

    assert count == len(filtered_actor_issues)
    assert isinstance(model, EqualGainModel)



from decide.data import database as db


def test_eq():
    """
    Shows the possibility of the hash/dictionary key implement
    """
    store = {}

    data_set = db.DataSet(name="pizza")
    store[data_set] = data_set

    assert store["pizza"] == data_set

    iteration = db.Iteration()
    iteration.pointer = 1

    iteration2 = db.Iteration()
    iteration2.pointer = 2

    iteration3 = db.Iteration()
    iteration3.pointer = 1

    store[iteration3] = iteration3

    assert iteration2 != iteration
    assert iteration3 == iteration
    assert store[iteration3] == iteration
    assert store[1] == iteration


def test_database(sqlite_db):
    created = db.DataSet.create(name="test123")

    data_set = db.DataSet.get(name="test123")

    assert data_set == created


def test_exchange_database():
    db.connection.drop_tables([db.ExchangeActor, db.Exchange])
    db.connection.create_tables([db.ExchangeActor, db.Exchange])



from typing import Iterable


def get_all_actors(connection, model_run_ids=None, data_set_id=None):
    assert not (model_run_ids and data_set_id), 'Provide only one of both arguments'

    if model_run_ids:
        sql = get_all_actors_by_model_run_ids_sql(model_run_ids)
    else:
        sql = get_all_actors_by_data_set_id_sql(data_set_id)

    cursor = connection.execute(sql)

    return cursor.fetchall()


def get_all_actors_by_model_run_ids_sql(model_run_ids):
    assert isinstance(model_run_ids, Iterable), 'model_run_ids should be an list of integers'

    model_run_id_param = list_to_sql_param(model_run_ids)

    sql = """SELECT a.name, a.id
    FROM actor a
             INNER JOIN dataset d on a.data_set_id = d.id
             INNER JOIN modelrun m on d.id = m.data_set_id
    WHERE m.id in (%s)
    GROUP BY a.name, a.id 
    ORDER BY a.name""" % model_run_id_param

    return sql


def get_all_actors_by_data_set_id_sql(data_set_id):
    assert isinstance(data_set_id, int), 'data_set_id should be an integer '

    sql = """SELECT a.name, a.id
    FROM actor a
    WHERE a.data_set_id = %s
    GROUP BY a.name, a.id""" % str(data_set_id)

    return sql


def list_to_sql_param(list_object):
    """
    In some way the database API of python does not handle IN(?) very well.
    Therefor we need to create a string of our values
    :param list_object:'
    :return:
    """
    return ','.join(list(map(lambda x: "\'{}\'".format(x), list_object)))



from setuptools import setup, find_packages

from decide import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="decide-exchange-model",
    version=__version__,
    packages=find_packages(),
    url="https://github.com/foarsitter/decide-exchange-model",
    license="GPL-3.0",
    author="jelmert",
    author_email="info@jelmert.nl",
    description="Model of collective decision making",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": ["decide-cli=decide.cli:main"],
        "gui_scripts": ["decide-gui=decide.gui:main"],
    },
    # data_files=[('data/input', ['data/input/kopenhagen.csv', 'data/input/CoP21.csv'])],
    python_requires=">=3.6",
    install_requires=[
        "blinker==1.4",
        "environs==6.0.0",
        "matplotlib==3.0.2",
        "requests==2.21.0",
        "typesystem==0.2.2"
    ],
)



import os
from typing import List

from decide.data.utils import write_exchanges
from .. import base
from ..observers import observer


class ExchangesWriter(observer.Observer):
    """
    There are three stages of externalities
    By exchange, by issue set and by actor
    """

    def __init__(
        self, observable: observer.Observable, summary_only=False, before=False
    ):
        super().__init__(observable)
        self.summary_only = summary_only
        self.before = before

    def _create_directory(self, repetition: int):
        if not os.path.exists(
            "{0}/exchanges/{1}".format(self.output_directory, repetition)
        ):
            os.makedirs("{0}/exchanges/{1}".format(self.output_directory, repetition))

        if (
            not os.path.exists(
                "{0}/exchanges/{1}/initial".format(self.output_directory, repetition)
            )
            and self.before
        ):
            os.makedirs(
                "{0}/exchanges/{1}/initial".format(self.output_directory, repetition)
            )

    def before_iterations(self, repetition):
        self._create_directory(repetition)

    def before_loop(self, iteration: int, repetition: int = None):
        """
        Writes all the possible exchanges to the filesystem
        :param iteration: 
        :param repetition:`
        """

        if not self.before:
            return

        self.model_ref.sort_exchanges()

        salt = self._get_salt

        write_exchanges(
            "{0}/exchanges/{2}/initial/before.{1}.{2}.csv".format(
                self.output_directory, iteration, repetition, salt
            ),
            self.model_ref.exchanges,
        )

    def after_loop(
        self, realized: List[base.AbstractExchange], iteration: int, repetition: int
    ):
        """
        Writes al the executed exchanges to the filesystem
        :param repetition: 
        :param realized: 
        :param iteration:
        """

        salt = self._get_salt

        write_exchanges(
            filename="{0}/exchanges/{2}/round.{1}.{3}.csv".format(
                self.output_directory, iteration + 1, repetition, salt
            ),
            realized=realized,
        )

    @property
    def _get_salt(self):
        model_name = "random"
        from decide.model.equalgain import EqualGainModel

        if isinstance(self.model_ref, EqualGainModel):
            model_name = "equal"
            if self.model_ref.randomized_value is not None:
                model_name += "-" + str(round(self.model_ref.randomized_value, 2))

        return model_name
